#!/usr/bin/python


import subprocess, time
import tempfile
from shutil import which
import warnings
import os

#  On import check dependencies
dependencies = [which('RNAfold') is not None,
                which('RNAsubopt') is not None,
                which('RNAeval') is not None]
if False in dependencies:
    warnings.warn('RBS Calculator Vienna is missing dependency ViennaRNA!')
#  End dependency check

debug=0

#Class that encapsulates all of the functions from NuPACK 2.0
class ViennaRNA(dict):

    debug_mode = 0
    RT = 0.61597 #gas constant times 310 Kelvin (in units of kcal/mol)
    param_file = f"-P {os.path.dirname(os.path.realpath(__file__))}/rna_turner1999.par "


    def __init__(self,Sequence_List,material = "rna37"):

        self.ran = 0

        import re
        import random
        import string

        exp = re.compile('[ATGCU]',re.IGNORECASE)

        for seq in Sequence_List:
            if exp.match(seq) == None:
                error_string = "Invalid letters found in inputted sequences. Only ATGCU allowed. \n Sequence is \"" + str(seq) + "\"."
                raise ValueError(error_string)

        self["sequences"] = Sequence_List
        self["material"] = material
        self.install_location = os.path.dirname(os.path.realpath(__file__))

        random.seed(time.time())

        with tempfile.NamedTemporaryFile(delete=False) as temp_file_maker:
            self.prefix = temp_file_maker.name

    def mfe(self, strands, constraints, Temp , dangles, outputPS = False):

        self["mfe_composition"] = strands
        
        Temp = float(Temp)
        if Temp <= 0: raise ValueError("The specified temperature must be greater than zero.")

        seq_string = "&".join(self["sequences"])
        constraints = None
        if constraints is None:
            input_string = seq_string + "\n" 
        else: 
            input_string = seq_string + "\n" + constraints + "\n"

        #write sequence file 
        handle = open(self.prefix,"w")
        handle.write(input_string)
        handle.close()

        #Set arguments
        material = self["material"]

        if material == 'rna1999':
            param_file = f"-P {self.install_location}/rna_turner1999.par "
        else:
            param_file = ''

        if dangles is "none":
            dangles = " -d0 "
        elif dangles is "some":
            dangles = " -d1 "
        elif dangles is "all":
            dangles = " -d2 "

        if outputPS:
            outputPS_str = " "
        else:
            outputPS_str = " --noPS "
            
        if constraints is None:
            args = outputPS_str + dangles + param_file + self.prefix

        else:
            args = outputPS_str + dangles + "-C " + param_file + self.prefix


        #Call ViennaRNA C programs
        cmd = "RNAfold"
        #print(cmd + args)

        output = subprocess.Popen(cmd + args, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines = True) #newlines argument was added because subprocess.popen was appending b's to the list output (byte stuff? I dont totally understand)
        std_out = output.communicate()[0]
        #output.tochild.write(input_string)

        while output.poll() is None:
            try:
                output.wait()
                time.sleep(0.001)
            except:
                break

        #if debug == 1: print(output.stdout.read())

        #Skip the unnecessary output lines
        #line = output.stdout.read()
        #print(std_out)

        line = std_out
        words = line.split()
        #print("These are the mfe value for " + str(words))
        bracket_string = words[1]
        #print(bracket_string)
        #print("This is the bracket string " + str(bracket_string))
        (strands,bp_x, bp_y) = self.convert_bracket_to_numbered_pairs(bracket_string)
        #print(strands, bp_x, bp_y)

        energy = float(words[len(words)-1].replace(")","").replace("(","").replace("\n",""))

        self["program"] = "mfe"
        self["mfe_basepairing_x"] = [bp_x]
        self["mfe_basepairing_y"] = [bp_y]
        self["mfe_energy"] = [energy]
        self["totalnt"]=strands

        #print "Minimum free energy secondary structure has been calculated."

    def subopt(self, strands, constraints, energy_gap, Temp = 37.0, dangles = "some", outputPS = False):

        self["subopt_composition"] = strands

        if Temp <= 0: raise ValueError("The specified temperature must be greater than zero.")

        material = self["material"]
        if material == 'rna1999':
            param_file = f"-P {self.install_location}/rna_turner1999.par "
        else:
            param_file = ''

        seq_string = "&".join(self["sequences"])
        #seq_string = self["sequences"][0]

        #print(f'SEQ STRING: {seq_string}')
        if constraints is None:
            input_string = seq_string + "\n" 
        else: 
            input_string = seq_string + "\n" + constraints + "&........." "\n"

        #print(self.prefix)
        with open(self.prefix + '.txt', 'w') as handle:
            handle.write(input_string)

        #Set arguments
        material = self["material"]

        if dangles is "none":
            dangles = " -d0 "
        elif dangles is "some":
            dangles = " -d1 "
        elif dangles is "all":
            dangles = " -d2 "

        if outputPS:
            outputPS_str = ""
        else:
            outputPS_str = " -noPS "
        #Call ViennaRNA C programs

        cmd = "RNAsubopt"
        energy_gap = energy_gap + 2.481
        if constraints is None:
            args = " -e " + str(energy_gap) + " -T " + str(Temp) + dangles + "--sorted --en-only " + param_file + " < " + self.prefix + '.txt'
        else:
            args = " -e " + str(energy_gap) + " -T " + str(Temp) + dangles + "--sorted --en-only " + "-C " + param_file + " < " + self.prefix + '.txt'

        #print(self.prefix)
        #print(cmd + args)
        output = subprocess.Popen(cmd + args, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines = True) #newlines argument was added because subprocess.popen was appending b's to the list output (byte stuff? I dont totally understand)
        std_out = output.communicate()[0]
        #output.tochild.write(input_string)

        while output.poll() is None:
            try:
                output.wait()
                time.sleep(0.001)
            except:
                break

        if debug == 1: print(output.stdout.read())

        #Skip unnecessary line
        #line = output.std_out.readline()
        line = std_out

        self["subopt_basepairing_x"] = []
        self["subopt_basepairing_y"] = []
        self["subopt_energy"] = []
        self["totalnt"]=[]

        counter=0  # TODO: Test to see if the latest version of VIENNA requires this workaround
        #print(f"Line: {line}")
        findings = line.split("\n")
        findings = findings[1:]
        findings = [x.split() for x in findings]
        #print(findings)
        for finding in findings:
            if len(finding) != 2:
                continue
            if len(self["sequences"]) > 1:
                binding_positions = finding[0]
                binding_positions = binding_positions.split("&")
                if ")" not in binding_positions[1] and "(" not in binding_positions[1]:
                    #print("This binding site is fake")
                    continue
                else:
                    self["subopt_energy"].append(float(finding[1]))
                    (strands, bp_x, bp_y) = self.convert_bracket_to_numbered_pairs(finding[0])
                    #print(strands, bp_x, bp_y)
                    self["subopt_basepairing_x"].append(bp_x)
                    self["subopt_basepairing_y"].append(bp_y)
            else:
                self["subopt_energy"].append(float(finding[1]))
                (strands, bp_x, bp_y) = self.convert_bracket_to_numbered_pairs(finding[0])
                self["subopt_basepairing_x"].append(bp_x)
                self["subopt_basepairing_y"].append(bp_y)


        self["subopt_NumStructs"] = counter

        self["program"] = "subopt"

        #print "Minimum free energy and suboptimal secondary structures have been calculated."

    def energy(self, strands, base_pairing_x, base_pairing_y, Temp = 37.0, dangles = "all"):

        self["energy_composition"] = strands

        if Temp <= 0: raise ValueError("The specified temperature must be greater than zero.")
        seq_string = "&".join(self["sequences"])
        strands = [len(seq) for seq in self["sequences"]]
        bracket_string = self.convert_numbered_pairs_to_bracket(strands,base_pairing_x,base_pairing_y)
        input_string = seq_string + "\n" + bracket_string + "\n"

        #print(seq_string,strands, bracket_string, input_string)

        handle = open(self.prefix,"w")
        handle.write(input_string)
        handle.close()

        #Set arguments
        material = self["material"]
        if dangles is "none":
            dangles = " -d0 "
        elif dangles is "some":
            dangles = " -d1 "
        elif dangles is "all":
            dangles = " -d2 "

        if material == 'rna1999':
            param_file = f"-P {self.install_location}/rna_turner1999.par "
        else:
            param_file = ''

        #Call ViennaRNA C programs
        cmd = "RNAeval"
        args = dangles + param_file + self.prefix
        #print(cmd + args)

        output = subprocess.Popen(cmd + args, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines = True) #newlines argument was added because subprocess.popen was appending b's to the list output (byte stuff? I dont totally understand)


        while output.poll() is None:
            try:
                output.wait()
                time.sleep(0.001)
            except:
                break

        #if debug == 1: print output.fromchild.read()
        std_out = output.communicate()[0]
        #print(std_out)

        self["energy_energy"] = []        

        line = std_out
        words = line.split()
        #print(line, words)
        if not words:
            raise ValueError('Could not catch the output of RNAeval. Is Vienna installed and in your PATH?')
        energy = float(words[-1].replace("(","").replace(")","").replace("\n",""))

        self["program"] = "energy"
        self["energy_energy"].append(energy)
        self["energy_basepairing_x"] = [base_pairing_x]
        self["energy_basepairing_y"] = [base_pairing_y]

        return energy

    def convert_bracket_to_numbered_pairs(self,bracket_string):

        all_seq_len = len(bracket_string)
        #print(all_seq_len)
        bp_x = []
        bp_y = []
        strands = []

        for y in range(bracket_string.count(")")):
            bp_y.append([])
            #print(bp_y)

        last_nt_x_list = []
        counter=0
        num_strands=0
        #print(bracket_string)
        for (pos,letter) in enumerate(bracket_string[:]):
            if letter is ".":
                counter += 1

            elif letter is "(":
                bp_x.append(pos-num_strands)
                last_nt_x_list.append(pos-num_strands)
                counter += 1

            elif letter is ")":
                nt_x = last_nt_x_list.pop() #nt_x is list of "(" except last entry
                #print('this is the last_nt_x_list ' + str(last_nt_x_list.pop()))
                nt_x_pos = bp_x.index(nt_x) 
                bp_y[nt_x_pos] = pos-num_strands
                counter += 1

            elif letter is "&":
                strands.append(counter)
                counter=0
                num_strands+=1

            else:
                print("Error! Invalid character in bracket notation.")
    

        if len(last_nt_x_list) > 0:
            print("Error! Leftover unpaired nucleotides when converting from bracket notation to numbered base pairs.")

        strands.append(counter)
        if len(bp_y) > 1:
            bp_x = [pos+1 for pos in bp_x[:]] #Shift so that 1st position is 1
            bp_y = [pos+1 for pos in bp_y[:]] #Shift so that 1st position is 1
            #print("subopt bp_x " + str(bp_x))
            #print("subopt bp_y " + str(bp_y))
        
        return (strands,bp_x, bp_y)

    def convert_numbered_pairs_to_bracket(self,strands,bp_x,bp_y):

        bp_x = [pos-1 for pos in bp_x[:]] #Shift so that 1st position is 0
        bp_y = [pos-1 for pos in bp_y[:]] #Shift so that 1st position is 0

        bracket_notation = []
        counter=0
        for (strand_number,seq_len) in enumerate(strands):
            if strand_number > 0: bracket_notation.append("&")
            for pos in range(counter,seq_len+counter):
                if pos in bp_x:
                    bracket_notation.append("(")
                elif pos in bp_y:
                    bracket_notation.append(")")
                else:
                    bracket_notation.append(".")
            counter+=seq_len

        return "".join(bracket_notation)

'''
if __name__ == "__main__":

    sequences = ["TCTGGCAGGGACCTGCACACGGATTGTGTGTGTTCCAGAGATGATAAAAAAGGAGTTAGTCTTGGTATGAGTAAAGGAGAAGAACTTTTCACTGGAGTTGTCCCAATTCTTGTTGAATTAGATGGTGATGTTAATGGGCACAAATTTTCTGTCAGTGGAGAGGGTG"] #,"acctcctta"]
    len(sequences) #,"acctcctta"]
    const_str = None

    test = ViennaRNA(sequences, material = "rna37")
    print(test)
    #dangles = "none"
    #constraints = none

    test.mfe([1], constraints = const_str , dangles = "none", Temp = 37.0)

    bp_x = test["mfe_basepairing_x"][0]
    bp_y = test["mfe_basepairing_y"][0]
    strands = test["totalnt"]
    bracket_string = test.convert_numbered_pairs_to_bracket(strands,bp_x,bp_y)
    print(bracket_string)

    (strands,bp_x, bp_y) = test.convert_bracket_to_numbered_pairs(bracket_string)

    print("Strands = ", strands)
    print("bp_x = ", bp_x)
    print("bp_y = ", bp_y)

    print(test.energy(strands, bp_x, bp_y, dangles = "all"))
    test.subopt(strands,const_str,0.5,dangles = "all")
    print(test)

#    print bracket_string
#    print test.convert_numbered_pairs_to_bracket(strands,bp_x,bp_y)
'''


import subprocess
from ostir import ostir
from shutil import which
import concurrent.futures
import json
import csv
import os



expected_vienna_version = '2.4.15'

class ValidationError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


def generate_validation_table():
    RNAs = []
    starts = []
    csv_keys = []
    file_location = os.path.dirname(os.path.realpath(__file__))
    with open(f'{file_location}/salis_2009_input.csv', 'r') as csv_input:
        reader = csv.reader(csv_input, delimiter=',')
        for i1, row in enumerate(reader):
            if i1 == 0:
                csv_keys = row
            else:
                single_input = {}
                for i2, item in enumerate(row):
                    if item == '':
                        item = None
                    if csv_keys[i2] == 'seq':
                        item = item.replace(' ', '')
                        RNAs.append(item)
                    elif csv_keys[i2] == 'start':
                        if item == '':
                            item = None
                        else:
                            item = int(item)
                        starts.append(item)

    input_sequences = list(zip(RNAs, starts))



    def _get_rbs_details(data_in):
        seq, start = data_in
        findings, details = ostir.run_ostir(seq, start_loc=start, detailed_out=True)
        return seq, list(findings), details

    with concurrent.futures.ThreadPoolExecutor(1) as multiprocessor:
        results = multiprocessor.map(_get_rbs_details, input_sequences)

    results = [x for x in results]

    result_table = {}
    for result in results:
        result_table[result[0]] = {'findings': result[1],
                                   'details': result[2]}
    with open(f'{file_location}/ostir_validation_table.json', 'w') as outfile:
        json.dump(result_table, fp=outfile, indent=4)



def verify_ostir_install():
    file_location = os.path.dirname(os.path.realpath(__file__))

    # Check for missing critical ViennaRNA executables
    dependencies = [which('RNAfold') is not None,
                    which('RNAsubopt') is not None,
                    which('RNAeval') is not None]
    if False in dependencies:
        raise ValidationError(f'Vienna installation was missing critical programs. Is ViennaRNA installed?')

    #  Check ViennaRNA version
    vienna_version = subprocess.check_output(['RNAfold', '--version'])
    vienna_version = str(vienna_version.strip()).replace("'", "").split(' ')[1]
    if vienna_version != expected_vienna_version:
        raise ValidationError(f'Vienna version did not match {expected_vienna_version} (was {vienna_version}).')

    # Check to make sure predefined RNAs match expected
    with open(f'{file_location}/ostir_validation_table.json', 'r') as infile:
        validation_table = json.load(infile)
    multiprocessing_in = [[index, value['details'], value['findings']] for index, value in validation_table.items()]

    def _verify_predefined_RNAs(table_info):
        findings, details = ostir.run_ostir(table_info[0], detailed_out=True)
        if table_info[2] == details and table_info[1] == findings:
            return False
        else:
            return True

    with concurrent.futures.ThreadPoolExecutor() as multiprocessor:
        result = multiprocessor.map(_verify_predefined_RNAs, multiprocessing_in)
    result = [x for x in result]

    if False in result:  # If true, validation has failed
        raise ValidationError(f'At least 1 RNA test did not match expected.')

    print('OSTIR validation successful')

if __name__ == '__main__':
    verify_ostir_install()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="osTIR",
    version="0.0.1",
    author="Barrick Lab/croots",
    author_email="croots@utexas.edu",
    description="Open Source Transcription Initiation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/barricklab/rbs-calculator",
    packages=setuptools.find_packages(exclude="tests"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    include_package_data=True,
    entry_points='''
        [console_scripts]
        ostir=ostir.ostir:main
        '''
)


#!/usr/bin/env python

import argparse
import math
from ostir.ostir_factory import OSTIRFactory
import os
from pathlib import Path
import subprocess
import concurrent.futures
import itertools
import csv
import copy

ostir_version = '0.0.1 (In-Development)'

def calc_dG_from_file(handle, output, verbose=True, parameters={}):
    from Bio import SeqIO

    records = SeqIO.parse(handle, "fasta")
    First = True
    export_PDF = True

    # for i in range(30):
    #    records.next()

    for record in records:

        mRNA = record.seq.tostring().upper()

        # Set any defaults
        start_range = [0, len(mRNA)]
        name = record.description.split(" ")[0]

        # Create instance of RBS Calculator
        test = OSTIRFactory(mRNA, start_range, name)

        # Examine kvars dictionary and pull out any options. Assign them to instanced class.
        for (key, value) in list(parameters.items()):

            if key == "cutoff":
                test.cutoff = value
            elif key == "start_range":
                test.start_range = value
            elif key == "rRNA":
                test.rRNA = value
            elif key == "energy_cutoff":
                test.energy_cutoff = value
            elif key == "standby_site_length":
                test.standby_site_length = value
            elif key == "dangles":
                test.dangles = value
            elif key == "export_PDF":
                export_PDF = value

        test.calc_dG()
        test.print_dG(test.infinity, print_expression=verbose)
        test.save_data(output, First)

        if First:
            First = False

        if export_PDF:
            num_structs = len(test.mRNA_rRNA_uncorrected_structure_list)
            for (structure, counter) in zip(test.mRNA_rRNA_uncorrected_structure_list, list(range(num_structs))):
                index = structure["MinStructureID"]
                structure.export_PDF(index, name, filename=name + "_rRNA" + "_" + str(counter) + ".pdf",
                                     program="subopt")

            num_structs = len(test.mRNA_structure_list)
            for (structure, counter) in zip(test.mRNA_structure_list, list(range(num_structs))):
                structure.export_PDF(0, name, filename=name + "_mRNA" + "_" + str(counter) + ".pdf")

    output.close()


def calc_dG_pre_post_RBS(pre_list, post_list, RBS_list, name_list, output, verbose=True, parameters={}):

    First = True

    for (pre, post, RBS, name) in zip(pre_list, post_list, RBS_list, name_list):

        mRNA = pre + RBS + post

        start_range = [0, len(mRNA)]

        # Create instance of RBS Calculator
        test = OSTIRFactory(mRNA, start_range, name)

        # Examine kvars dictionary and pull out any options. Assign them to instanced class.
        for (key, value) in list(parameters.items()):

            if key == "cutoff":
                test.cutoff = value
            elif key == "start_range":
                test.start_range = value
            elif key == "rRNA":
                test.rRNA = value
            elif key == "energy_cutoff":
                test.energy_cutoff = value
            elif key == "standby_site_length":
                test.standby_sitRBSe_length = value
            elif key == "dangles":
                test.dangles = value
            elif key == "export_PDF":
                export_PDF = value

        test.calc_dG()
        if verbose: test.print_dG(test.infinity)

        test.save_data(output, First)
        if First:
            First = False

    output.close()


def run_ostir(seq, constraint_str=None, outfile=None, start_loc=0, end_loc=None, i=None, verbose=False,
                          detailed_out=False, print_out=False, sd=None):
    '''Takes an RNA with optional paramaters and returns binding energies.

        Keyword arguments:
        seq -- Sequence to calculate binding energies for
        constraint_str --
        outfile -- Filepath for output csv
        start_loc -- First base to start considering start codons. Defaults to first base
        end_loc -- Last base to start considering start codons. Defaults to end of sequence
        i -- Returns i as part of the first return variable, useful for tagging things for downstream processing.
        verbose -- Prints debug information
        detailed_out -- returns components of total dG as an additional return variable
        print_out -- prints output to the console
        sd -- Defines anti-Shine-Dalgarno sequence. Defaults to that of E. coli's

    '''
    mRNA = seq
    if end_loc == None:
        end_loc = len(mRNA)
    start_range = [start_loc, end_loc]

    if not sd:
        sd = 'ACCTCCTTA'

    calcObj = OSTIRFactory(mRNA, start_range, sd, constraint_str, verbose=verbose)
    calcObj.calc_dG()

    dG_total_list = calcObj.dG_total_list[:]
    dG_details = calcObj.dG_details[:]
    start_pos_list = calcObj.start_pos_list[:]
    kinetic_score_list = calcObj.kinetic_score_list[:]
    standby_site_list = calcObj.dG_standby_site_list[:]
    start_codon_list = calcObj.start_codon_list[:]

    expr_list = []
    for dG in dG_total_list:
        expr_list.append(calcObj.K * math.exp(-dG / calcObj.RT_eff))

    return_var = zip(expr_list, start_pos_list, kinetic_score_list, dG_total_list, standby_site_list, start_codon_list)

    if outfile:
        with open(outfile, 'w') as out_file:
            for (expr, start_pos, ks, dG, dstandby, codon) in zip(expr_list, start_pos_list, kinetic_score_list, dG_total_list,
                                                           standby_site_list, start_codon_list):
                out_file.writelines(['1\n', f"{start_pos} {expr} {ks}, {i}\n"])
        if detailed_out:
            second_out = outfile + '.detailed'
            with open(second_out, 'w') as outfile:
                pass
    if print_out:
        calcObj.print_dG()
    if detailed_out:
        return return_var, dG_details
    else:
        return return_var

def parse_fasta(filepath):
    '''Takes a filepath to a fasta formatted file and returns a list of [header, sequence].'''
    sequences = []
    current_seq_name = None
    current_seq = ""
    with open(filepath, 'r') as infile:
        for line in infile:
            linestr = str(line)
            if linestr[0] == '>':
                if current_seq_name:
                    sequences.append([current_seq_name, current_seq])
                current_seq_name = linestr[1:].rstrip()
                current_seq = str()
                continue
            else:
                current_seq += linestr.rstrip()
    if current_seq_name:
        sequences.append([current_seq_name, current_seq])
    return sequences

def _parallelizer(input):
    '''Used to process concurrent.futures input into run_ostir(), and processes the output for csv/command line out.'''

    def _valid_entry(key_to_test):
        '''Returns False if command line/csv input was empty, else returns true.'''
        if key_to_test not in input.keys():
            return False
        elif input[key_to_test] == '':
            return False
        elif not input[key_to_test]:
            return False
        else:
            return True

    #unpack args:
    seq = input['seq']
    if 'constraint_str' in input.keys():
        constraint_str = input['constraint_str']
    else:
        constraint_str = None
    if 'outfile' in input.keys():
        outfile = input['outfile']
    else:
        outfile = None
    if _valid_entry('start'):
        start_loc = int(input['start'])-1
    else:
        start_loc = 0
    if _valid_entry('end'):
        end_loc = int(input['end'])-1
    elif _valid_entry('start'):
        end_loc = int(input['start'])+1
    else:
        end_loc = len(seq)
    if 'i' in input.keys():
        i = input['i']
    else:
        i = None
    if 'verbose' in input.keys():
        verbose = input['verbose']
    else:
        verbose = None
    if 'detailed_out' in input.keys():
        detailed_out = input['detailed_out']
    else:
        detailed_out = None
    if 'print_out' in input.keys():
        print_out = input['print_out']
    else:
        print_out = False
    if 'sd' in input.keys():
        sd = input['sd']
    else:
        sd = None


    normal, detailed= run_ostir(seq, constraint_str, outfile, start_loc,
                                end_loc, i, verbose, detailed_out, print_out, sd)
    normal = list(normal)
    zip_output = zip(normal, detailed)
    output_data_list = []
    for output in list(zip_output):
        outdata = {'RNA': seq,
                   'codon': output[0][5],
                   'start_pos': output[0][1]+1,
                   'dG_total': output[0][3],
                   'dG_rRNA:mRNA': output[1][0],
                   'dG_mRNA': output[1][2],
                   'dG_Spacing': output[1][4],
                   'dG_Standby': output[1][3],
                   'dG_Start_Codon': output[1][1],
                   'Expression': output[0][0]
                   }
        output_data_list.append(outdata)
    return output_data_list

def _print_output(outlist):
    '''Processes command line / csv input for pretty command line output'''
    sorted_predictions = {}
    keys = []
    for prediction in outlist:
        if prediction['RNA'] in sorted_predictions.keys():
            sorted_predictions[prediction['RNA']].append(prediction)
        else:
            sorted_predictions[prediction['RNA']] = [prediction]
            keys.append(prediction['RNA'])

    output_items = ['start_pos', 'codon', 'Expression', 'dG_total', 'dG_rRNA:mRNA', 'dG_mRNA', 'dG_Spacing', 'dG_Standby', 'dG_Start_Codon']
    row_format = "{:>15}" * (len(output_items))
    print('_________________________________________________')
    for rna in keys:
        print(f'Tested Sequence: {rna}')
        print(row_format.format(*output_items))
        for start in sorted_predictions[rna]:
            output_data = [start[key] for key in output_items]
            for i, data_point in enumerate(output_data):
                if type(data_point) == float:
                    data_point = format(data_point, '.4f')
                    output_data[i] = data_point
            print(row_format.format(*output_data))
        print('_________________________________________________')
        pass


def main():
    '''Main OSTIR function ran when called as an executable. See ostir -h.'''
    parser = argparse.ArgumentParser(description='Open Source Transcription Initiation Rates')

    parser.add_argument(
        '-i', '--input',
        action='store',
        metavar='str/filepath',
        dest='i',
        required=False,
        type=str,
        help="input DNA/RNA. Required if not using --validate.",
    )

    parser.add_argument(
        '-o', '--output',
        action='store',
        metavar= 'filepath',
        dest='o',
        required=False,
        type=str,
        help="Output filepath. If not provided, results will output to the console.",
    )

    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        dest='v',
        required=False,
        help="Prints verbose output.",
    )

    parser.add_argument(
        '-s', '--start',
        action='store',
        metavar='int',
        dest='s',
        required=False,
        type=int,
        help="Defines most 5' position to consider start codons.",
    )

    parser.add_argument(
        '-e', '--end',
        action='store',
        metavar='int',
        dest='e',
        required=False,
        type=int,
        help="Defines most 3' position to consider start codons.",
    )

    parser.add_argument(
        '-r', '--rRNA',
        action='store',
        metavar='str',
        dest='r',
        required=False,
        type=int,
        help="Defines rRNA anti-Shine-Dalgarno sequence. Defaults to that of E. coli's.",
    )

    parser.add_argument(
        '-j', '--threads',
        action='store',
        metavar='int',
        dest='j',
        required=False,
        type=int,
        help="Number of threads for multiprocessing",
    )

    parser.add_argument(
        '--validate',
        action='store_true',
        dest='validate',
        required=False,
        help="Runs a consistency test to ensure proper installation.",
    )

    options = parser.parse_args()

    if options.validate:
        from ostir.ostir_validation import verify_ostir_install
        verify_ostir_install()
        exit()
    elif not options.i:
        subprocess.run(['ostir', '-h'])
        exit(1)

    cmd_kwargs = dict()
    cmd_kwargs['seq'] = options.i
    cmd_kwargs['detailed_out'] = True
    if options.o:
        outfile = options.o
    else:
        outfile = None
        cmd_kwargs['print_out'] = False
    if options.j:
        cores = options.j
    else:
        cores = 1
    if options.v:
        cmd_kwargs['verbose'] = options.v
    if options.s:
        cmd_kwargs['start'] = options.s
    if options.e:
        cmd_kwargs['end'] = options.e
    if options.r:
        cmd_kwargs['sd'] = options.r

    vienna_version = subprocess.check_output(['RNAfold', '--version'])
    vienna_version = str(vienna_version.strip()).replace("'", "").split(' ')[1]
    print(f'Running osTIR Version {ostir_version} (with Vienna version: {vienna_version})')

    # Output data: RNA, Codon, position, dg_total, dg rRNA:mRNA, dg mRNA, dG Spacing, dg Standby, Kinetic Score

    if os.path.isfile(cmd_kwargs['seq']):
        if Path(cmd_kwargs['seq']).suffix == '.fasta':
            sequences = parse_fasta(cmd_kwargs['seq'])
            concurrent_futures_args = []
            for sequence in sequences:
                cmd_kwargs['seq'] = sequence[1]
                concurrent_futures_args.append(copy.deepcopy(cmd_kwargs))
            with concurrent.futures.ThreadPoolExecutor(max_workers=cores) as multiprocessor:
                result = multiprocessor.map(_parallelizer, concurrent_futures_args)
            result = [x for x in result]
            result = list(itertools.chain.from_iterable(result))
            if outfile:
                csv_keys = result[0].keys()
                with open(outfile, 'w')  as output_file:
                    dict_writer = csv.DictWriter(output_file, csv_keys)
                    dict_writer.writeheader()
                    dict_writer.writerows(result)
            else:
                _print_output(result)
        elif Path(cmd_kwargs['seq']).suffix == '.csv':
            csv_keys = []
            csv_values = []
            with open(cmd_kwargs['seq'], 'r') as csv_input:
                reader = csv.reader(csv_input, delimiter=',')
                for i1, row in enumerate(reader):
                    if i1 == 0:
                        csv_keys = row
                    else:
                        single_input = {}
                        for i2, item in enumerate(row):
                            if item == '':
                                item = None
                            if csv_keys[i2] == 'seq':
                                item = item.replace(' ', '')
                            cmd_kwargs[csv_keys[i2]] = item
                        csv_values.append(copy.deepcopy(cmd_kwargs))
            with concurrent.futures.ThreadPoolExecutor(max_workers=cores) as multiprocessor:
                result = multiprocessor.map(_parallelizer, csv_values)
            result = [x for x in result]
            result = list(itertools.chain.from_iterable(result))
            if outfile:
                csv_keys = result[0].keys()
                with open(outfile, 'w')  as output_file:
                    dict_writer = csv.DictWriter(output_file, csv_keys)
                    dict_writer.writeheader()
                    dict_writer.writerows(result)
            else:
                _print_output(result)

        else:
            raise ValueError('Input file is not a supported filetype')
    else:
        result = _parallelizer(**cmd_kwargs)
        if outfile:
            csv_keys = result[0].keys()
            with open(outfile, 'w')  as output_file:
                dict_writer = csv.DictWriter(output_file, csv_keys)
                dict_writer.writeheader()
                dict_writer.writerows(result)
        else:
            _print_output(result)


if __name__ == "__main__":
    main()